package fse.eclipse.branchrecorder.changerecorder;

import java.io.File;
import java.io.IOException;
import java.nio.file.Path;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.commons.io.FileUtils;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IWorkspaceRoot;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.jgit.lib.Repository;
import org.jtool.changerecorder.history.OperationHistory;
import org.jtool.changerecorder.history.Xml2Operation;
import org.jtool.changerecorder.operation.IOperation;
import org.jtool.changerecorder.util.Whoami;
import org.jtool.changerecorder.util.XmlFileStream;
import org.w3c.dom.Document;

import fse.eclipse.branchrecorder.util.PathUtil;

public class OperationCopier {
    private Map<String, ProjectCache> cacheMap = new HashMap<>();
    private final IWorkspaceRoot wRoot = ResourcesPlugin.getWorkspace().getRoot();
    private final String parentDirPath;

    OperationCopier(String parentDirPath) {
        this.parentDirPath = parentDirPath;
    }

    /**
     * Historyファイルから，そのプロジェクトがGitで管理されているかを判定 <br>
     * <code>true</code>ならHistoryファイルをコピーする
     */
    public void copy(Path xmlFile) throws InterruptedException, IOException {
        String filePath = validateXmlFilePath(xmlFile);
        List<IOperation> operations = readOperations(filePath);
        String projectName = readProjectName(operations);
        if (projectName == null || projectName.length() == 0) {
            return;
        }

        ProjectCache cache = getProjectCache(projectName);
        if (cache != null && cache.hasRepository()) {
            String branchName = cache.getBranchName();
            copyHistoryFile(filePath, cache.getProject(), Whoami.getUserName(), branchName);
            // プロジェクトをrefreshして，コピーしたファイルを即時反映する
            cache.refreshIProject();
        }
    }

    private ProjectCache getProjectCache(String projectName) {
        ProjectCache cache = cacheMap.getOrDefault(projectName, null);
        if (cache == null) {
            IProject p = wRoot.getProject(projectName);
            Repository r = RepositoryCache.getRepository(p);
            if (p != null && r != null) {
                cache = new ProjectCache(p, r);
                cacheMap.put(projectName, cache);
            }
        }
        return cache;
    }

    private String validateXmlFilePath(Path xmlFile) {
        String fileName = xmlFile.toString();

        StringBuilder sb = new StringBuilder();
        sb.append(parentDirPath).append(File.separator).append(fileName);
        return sb.toString();
    }

    /**
     * Historyファイルから全Operationを取り出す
     */
    private List<IOperation> readOperations(String filePath) throws InterruptedException {
        // ファイルが作成された瞬間に飛んでくるため(?)，読み込みに失敗することがある(NullPointerException)
        try {
            Document doc = XmlFileStream.read(filePath);
            OperationHistory history = Xml2Operation.convert(doc);
            return history.getOperations();
        } catch (NullPointerException n) {
            // 読み込みに失敗した際，一瞬間をおき再度処理
            Thread.sleep(100);
            return readOperations(filePath);
        }
    }

    /**
     * Operationからプロジェクト名を抜き取る
     */
    private String readProjectName(List<IOperation> operations) {
        for (IOperation operation : operations) {
            String projPath = operation.getFilePath();
            if (!isNullPath(projPath)) {
                return cutoutProjectName(projPath);
            }
        }
        return "";
    }

    /**
     * Operationから読み込んだファイルパスが使用不可能なパスかどうか
     */
    private boolean isNullPath(String projPath) {
        return projPath == null || projPath.length() == 0 || projPath.equals("null");
    }

    /**
     * Operationのファイルパスからプロジェクト名を抜き取る
     */
    private String cutoutProjectName(String projPath) {
        int firstIndexOf = projPath.indexOf("/") + 1;
        int secondIndexOf = projPath.indexOf("/", firstIndexOf);
        return projPath.substring(firstIndexOf, secondIndexOf);
    }

    /**
     * Historyファイルを プロジェクトパス/#history/ブランチ名 へコピーする
     */
    private void copyHistoryFile(String filePath, IProject project, String author, String branchName) {
        File srcFile = new File(filePath);
        File destDir = new File(PathUtil.getProjectBranchPath(project, author, branchName));
        if (!destDir.exists()) {
            try {
                FileUtils.forceMkdir(destDir.getParentFile());
            } catch (IOException e) {
                e.printStackTrace();
            }
        }

        try {
            FileUtils.copyFileToDirectory(srcFile, destDir);
        } catch (IOException e) {
            e.printStackTrace();
        }
    }
}
