package fse.eclipse.mergehelper.detector;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import org.jtool.changerepository.data.FileInfo;
import org.jtool.changerepository.data.ProjectInfo;
import org.jtool.changerepository.dependencygraph.FileOpDepGraph;
import org.jtool.changerepository.dependencygraph.OpDepGraph;
import org.jtool.changerepository.dependencygraph.OpDepGraphInfo;
import org.jtool.changerepository.dependencygraph.OpDepGraphNode;
import org.jtool.changerepository.operation.UnifiedOperation;
import org.jtool.changerepository.parser.CodeRange;
import org.jtool.changerepository.parser.OpJavaElement;
import org.jtool.changeslicereplayer.slicer.CodeSnippet;
import org.jtool.changeslicereplayer.slicer.OpGraphSlicer;
import org.jtool.changeslicereplayer.slicer.Slice;
import org.jtool.changeslicereplayer.slicer.SliceCriterion;

import fse.eclipse.mergehelper.element.BranchFileInfo;
import fse.eclipse.mergehelper.element.BranchInfo;
import fse.eclipse.mergehelper.element.BranchJavaElement;
import fse.eclipse.mergehelper.element.BranchRootInfo;
import fse.eclipse.mergehelper.element.MergeType;
import fse.eclipse.mergehelper.util.Parser;
import fse.eclipse.mergehelper.util.RepositoryElementInfoUtil;

public class OperationAllocator extends AbstractDetector {
    private static final String TASK_NAME = "Allocate Operation";
    private static final String ERROR_MESSAGE = "NULL";
    private static AbstractDetector instance = new OperationAllocator();

    private OperationAllocator() {
    }

    public static AbstractDetector getInstace() {
        return instance;
    }

    @Override
    protected void execute() {
        BranchRootInfo rootInfo = BranchRootInfo.getInstance();

        if (DetectOption.getInstance().isSlice()) {
            allocateOperationBySlice(rootInfo.getBranchInfo(MergeType.ACCEPT));
            allocateOperationBySlice(rootInfo.getBranchInfo(MergeType.JOIN));
        } else {
            allocateOperationByParse(rootInfo.getBranchInfo(MergeType.ACCEPT));
            allocateOperationByParse(rootInfo.getBranchInfo(MergeType.JOIN));
        }
    }

    private void allocateOperationByParse(BranchInfo bInfo) {
        ProjectInfo pInfo = bInfo.getProjectInfo();
        List<FileInfo> fInfos = pInfo.getAllFileInfo();

        String branchName = bInfo.getName();
        int size = fInfos.size();

        for (int i = 0; i < size; i++) {
            sendSubMessage(branchName, i + 1, size);

            FileInfo fInfo = fInfos.get(i);
            if (RepositoryElementInfoUtil.hasEditOperation(fInfo)) {
                BranchFileInfo bfInfo = createBranchFileInfo(bInfo, fInfo);
                if (bfInfo != null) {
                    List<UnifiedOperation> ops = fInfo.getOperations();
                    int osize = ops.size();
                    for (int j = 0; j < osize; j++) {
                        String code = fInfo.getCode(j);
                        UnifiedOperation op = ops.get(j);
                        OpJavaElement elem = searchBranchJavaElement(fInfo, code, op);
                        if (elem != null) {
                            bfInfo.addOperation(elem, op);
                        }
                    }

                    removeNotEditedElement(bfInfo);
                }
            }
        }
    }

    private BranchFileInfo createBranchFileInfo(BranchInfo bInfo, FileInfo fInfo) {
        int size = fInfo.getOperationNumber() - 1;
        for (int i = size; i >= 0; i--) {
            String latestCode = fInfo.getCode(i);
            List<OpJavaElement> elems = new ArrayList<>();
            boolean parseable = Parser.collectElements(fInfo, latestCode, elems, false);
            if (parseable) {
                return bInfo.createBranchFileInfo(fInfo, elems);
            }
        }
        return null;
    }

    private OpJavaElement searchBranchJavaElement(FileInfo fInfo, String code, UnifiedOperation op) {
        List<OpJavaElement> elems = new ArrayList<>();
        boolean parseable = Parser.collectElements(fInfo, code, elems, true);
        if (!parseable || elems.size() == 0) {
            return null;
        }

        int start = op.getStart();
        for (OpJavaElement elem : elems) {
            CodeRange range = elem.getCodeRange();
            if (range.inRange(start)) {
                return elem;
            }
        }
        return null;
    }

    private void allocateOperationBySlice(BranchInfo bInfo) {
        ProjectInfo pInfo = bInfo.getProjectInfo();
        OpDepGraphInfo.reset();
        OpDepGraph graph = OpDepGraphInfo.createGraph(pInfo);
        if (graph == null) {
            throw new Error();
        }

        List<FileOpDepGraph> fGraphs = graph.getFileGraphs();

        String branchName = bInfo.getName();
        int size = fGraphs.size();

        for (int i = 0; i < size; i++) {
            sendSubMessage(branchName, i + 1, size);

            FileOpDepGraph fGraph = fGraphs.get(i);
            FileInfo fInfo = fGraph.getFileInfo();

            int idx = fInfo.getOperationNumber() - 1;
            String code = fInfo.getCode(idx);
            List<OpJavaElement> elems = new ArrayList<>();
            boolean parseable = Parser.collectElements(fInfo, code, elems, false);
            if (!parseable) {
                throw new Error();
            }

            BranchFileInfo bfInfo = bInfo.createBranchFileInfo(fInfo, elems);
            List<Slice> slices = new ArrayList<>();
            Map<String, OpJavaElement> elemMap = new HashMap<>();
            long time = fInfo.getOperation(idx).getTime();
            for (OpJavaElement elem : elems) {
                List<CodeSnippet> snippets = new ArrayList<>();
                for (CodeRange range : elem.getRanges()) {
                    int start = range.getStart();
                    int end = range.getEnd();
                    String text = code.substring(start, end);
                    snippets.add(new CodeSnippet(start, end, idx, time, text));
                }

                String name = elem.getSimpleName();
                SliceCriterion criterion = new SliceCriterion(fGraph, name, snippets);
                Slice slice = OpGraphSlicer.constructBackwardSlice(criterion);
                OpDepGraph.sortNodes(slice.getNodes());
                adjustSliceNode(slice);
                slices.add(slice);
                elemMap.put(name, elem);
            }
            removeDuplicateOperationId(fInfo, slices);

            for (Slice slice : slices) {
                String name = slice.getCriterion().getName();
                OpJavaElement elem = elemMap.get(name);
                for (OpDepGraphNode node : slice.getNodes()) {
                    bfInfo.addOperation(elem, node.getOperation());
                }
            }
            removeNotEditedElement(bfInfo);
        }
    }

    private void adjustSliceNode(Slice slice) {
        int beforeId = -1;
        List<OpDepGraphNode> nodes = slice.getNodes();
        for (int i = 0; i < nodes.size(); i++) {
            OpDepGraphNode node = nodes.get(i);
            UnifiedOperation op = node.getOperation();
            if (op.isTextChangedOperation()) {
                int id = node.getId();
                if (beforeId != id) {
                    beforeId = id;
                } else {
                    nodes.remove(i);
                    i--;
                }
            } else {
                nodes.remove(i);
                i--;
            }
        }
    }

    private void removeDuplicateOperationId(FileInfo fInfo, List<Slice> slices) {
        if (slices.size() <= 1) {
            return;
        }

        List<UnifiedOperation> ops = fInfo.getOperations();
        List<Slice> duplicateSlices = new ArrayList<>();
        for (UnifiedOperation op : ops) {
            for (Slice slice : slices) {
                if (slice.contain(op)) {
                    duplicateSlices.add(slice);
                }
            }
            if (duplicateSlices.size() >= 2) {
                String code = fInfo.getCode(op.indexOfFileInfo());
                int id = op.getId();
                OpJavaElement elem = searchBranchJavaElement(fInfo, code, op);
                if (elem != null) {
                    String name = elem.getSimpleName();
                    for (Slice slice : duplicateSlices) {
                        if (!slice.getCriterion().getName().equals(name)) {
                            List<OpDepGraphNode> nodes = slice.getNodes();
                            for (OpDepGraphNode node : nodes) {
                                if (node.getOperation().getId() == id) {
                                    nodes.remove(node);
                                    break;
                                }
                            }
                        }
                    }
                } else {
                    for (Slice slice : duplicateSlices) {
                        List<OpDepGraphNode> nodes = slice.getNodes();
                        for (OpDepGraphNode node : nodes) {
                            if (node.getOperation().getId() == id) {
                                nodes.remove(node);
                                break;
                            }
                        }
                    }
                }
            }
            if (duplicateSlices.size() >= 1) {
                duplicateSlices.clear();
            }
        }
    }

    private void removeNotEditedElement(BranchFileInfo bfInfo) {
        Map<String, BranchJavaElement> elemMap = bfInfo.getElementMap();
        for (Entry<String, BranchJavaElement> entry : elemMap.entrySet()) {
            if (!entry.getValue().isEdited()) {
                bfInfo.removeBranchJavaElement(entry.getKey());
            }
        }
    }

    private void sendSubMessage(String branchName, int nowWork, int totalWork) {
        StringBuilder sb = new StringBuilder();
        sb.append(branchName).append("  ");
        sb.append(nowWork).append(" / ").append(totalWork);
        sendSubMessage(sb.toString());
    }

    @Override
    protected String getTaskName() {
        return TASK_NAME;
    }

    @Override
    protected String getErrorMessage() {
        return ERROR_MESSAGE;
    }

    @Override
    protected void nextState() {
        ConflictDetector.getInstance().detect(dialog);
    }
}
