package fse.eclipse.mergehelper.detector.merge;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import org.apache.commons.io.FileUtils;
import org.eclipse.jgit.api.Git;
import org.eclipse.jgit.api.errors.CheckoutConflictException;
import org.eclipse.jgit.api.errors.GitAPIException;
import org.eclipse.jgit.api.errors.InvalidRefNameException;
import org.eclipse.jgit.api.errors.RefAlreadyExistsException;
import org.eclipse.jgit.api.errors.RefNotFoundException;
import org.eclipse.jgit.lib.Ref;
import org.jtool.changerecorder.util.FileStream;
import org.jtool.changerepository.data.FileInfo;
import org.jtool.changerepository.data.ProjectInfo;
import org.jtool.changerepository.operation.UnifiedOperation;

import fse.eclipse.mergehelper.Activator;
import fse.eclipse.mergehelper.detector.AbstractDetector;
import fse.eclipse.mergehelper.detector.BranchHistoryCopier;
import fse.eclipse.mergehelper.element.BranchInfo;
import fse.eclipse.mergehelper.element.BranchRootInfo;
import fse.eclipse.mergehelper.element.MergePoint;
import fse.eclipse.mergehelper.element.MergeType;
import fse.eclipse.mergehelper.util.RepositoryElementInfoUtil;

public class AMerge extends AbstractDetector {
    private static final String TASK_NAME = "Do Artificial Merge ...";
    private static final String ERROR_MESSAGE = "Artificial Merge Failed";
    private static AbstractDetector instance = new AMerge();

    private boolean isSuccess;
    private List<MergeIndexHolder> mergeIdxHoloders;

    private AMerge() {
    }

    public static AbstractDetector getInstance() {
        return instance;
    }

    @Override
    public void execute() {
        BranchRootInfo rootInfo = BranchRootInfo.getInstance();
        BranchInfo a_bInfo = rootInfo.getBranchInfo(MergeType.ACCEPT);
        BranchInfo j_bInfo = rootInfo.getBranchInfo(MergeType.JOIN);
        MergePoint mPoint = rootInfo.getMergePoint();
        int size = mPoint.size();
        for (int i = 0; i < size; i++) {
            Git git = init(rootInfo);
            mergeIdxHoloders = new ArrayList<>();
            try {
                commit(git, mPoint.getOperation(i, MergeType.ACCEPT), a_bInfo);
                commit(git, mPoint.getOperation(i, MergeType.JOIN), j_bInfo);

                @SuppressWarnings("deprecation")
				Ref ref = git.getRepository().getRef(a_bInfo.getName());
                isSuccess = git.merge().include(ref).call().getMergeStatus().isSuccessful();
                if (isSuccess) {
                    mPoint.setReferenceIndex(i);
                    break;
                }
            } catch (GitAPIException | IOException e) {
                e.printStackTrace();
            }
        }
    }

    private Git init(BranchRootInfo rootInfo) {
        File dir = new File(Activator.getWorkingDirPath());
        String projectName = rootInfo.getName();
        try {
            Git git;
            if (dir.listFiles().length != 2) {
                FileUtils.deleteDirectory(new File(dir, ".git"));
                FileUtils.deleteDirectory(new File(dir, projectName));
            }
            Git.init().setDirectory(dir).call();
            git = Git.open(dir);

            BranchInfo a_bInfo = rootInfo.getBranchInfo(MergeType.ACCEPT);
            String branchName = a_bInfo.getName();

            ProjectInfo a_pInfo = a_bInfo.getProjectInfo();
            List<FileInfo> fInfos = a_pInfo.getAllFileInfo();
            for (FileInfo fInfo : fInfos) {
                String filePath = createMergePath(fInfo, branchName);
                FileStream.write(filePath, fInfo.getCode(0));
            }

            git.add().addFilepattern(projectName).call();
            git.commit().setMessage("MergeHelper-init").call();

            // master
            String nowBranchName = git.getRepository().getBranch();

            String a_branchName = rootInfo.getBranchName(MergeType.ACCEPT);
            String j_branchName = rootInfo.getBranchName(MergeType.JOIN);

            if (!nowBranchName.equals(a_branchName)) {
                git.branchCreate().setName(a_branchName).call();
            }

            if (!nowBranchName.equals(j_branchName)) {
                git.branchCreate().setName(j_branchName).call();
            }
            return git;
        } catch (IllegalStateException | GitAPIException | IOException e) {
            e.printStackTrace();
        }
        return null;
    }

    private void commit(Git git, UnifiedOperation op, BranchInfo bInfo)
            throws RefAlreadyExistsException, RefNotFoundException, InvalidRefNameException, CheckoutConflictException, GitAPIException {
        String branchName = bInfo.getName();
        git.checkout().setName(branchName).call();

        List<FileInfo> fInfos = bInfo.getProjectInfo().getAllFileInfo();
        for (FileInfo fInfo : fInfos) {
            String filePath = createMergePath(fInfo, branchName);
            int idx;
            if (op.getFileInfo().equals(fInfo)) {
                idx = fInfo.indexOfOperation(op.getId());
            } else {
                idx = RepositoryElementInfoUtil.indexOfJustBeforeFileOperation(fInfo, op);
            }
            String code = fInfo.getCode(idx);
            FileStream.write(filePath, code);
            mergeIdxHoloders.add(new MergeIndexHolder(fInfo, idx));
        }

        git.add().addFilepattern(BranchRootInfo.getInstance().getName()).call();
        git.commit().setMessage("MergeHelper-" + branchName).call();
    }

    private static String createMergePath(FileInfo fInfo, String branchName) {
        String filePath = fInfo.getFilePath();
        int idx = filePath.indexOf(BranchHistoryCopier.BRANCH_NAME_MARK);
        int idx2 = idx + BranchHistoryCopier.BRANCH_NAME_MARK.length() + branchName.length();

        StringBuilder sb = new StringBuilder(Activator.getWorkingDirPath());
        sb.append(filePath.substring(0, idx)).append(filePath.substring(idx2));
        return sb.toString();
    }

    @Override
    protected String getTaskName() {
        return TASK_NAME;
    }

    @Override
    protected String getErrorMessage() {
        return ERROR_MESSAGE;
    }

    @Override
    protected void nextState() {
        if (isSuccess) {
            AMergedAdjustOperationOffset.getInstance().detect(dialog, mergeIdxHoloders);
        } else {
            error();
        }
    }

    public static String getAMergedCode(FileInfo fInfo, String branchName) {
        String filePath = createMergePath(fInfo, branchName);
        String code = FileStream.read(filePath);
        return code;
    }

    public class MergeIndexHolder {
        private String name;
        private int idx;

        private MergeIndexHolder(FileInfo fInfo, int idx) {
            name = fInfo.getQualifiedName();
            this.idx = idx;
        }

        public int getIndex() {
            return idx;
        }

        public boolean equalsFileInfo(FileInfo fInfo) {
            return name.equals(fInfo.getQualifiedName());
        }
    }
}