package fse.eclipse.mergehelper.detector.merge;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.TreeMap;

import org.jtool.changerecorder.operation.NormalOperation;
import org.jtool.changerepository.data.FileInfo;
import org.jtool.changerepository.data.ProjectInfo;
import org.jtool.changerepository.operation.UnifiedOperation;

import fse.eclipse.mergehelper.detector.AbstractDetector;
import fse.eclipse.mergehelper.detector.DetectOption;
import fse.eclipse.mergehelper.element.BranchInfo;
import fse.eclipse.mergehelper.element.BranchJavaElement;
import fse.eclipse.mergehelper.element.BranchRootInfo;
import fse.eclipse.mergehelper.element.ConflictInfo;
import fse.eclipse.mergehelper.element.MergeType;
import fse.eclipse.mergehelper.util.RepositoryElementInfoUtil;

public class OperationSwapper extends AbstractDetector {
    private static final String TASK_NAME = "Swap Operation ...";
    private static final String ERROR_MESSAGE = "OperationSwap Debug Error  ";
    private static AbstractDetector instance = new OperationSwapper();

    private static final String UNKNOWN = "UNKNOWN";
    private static final String MERGEABLE = "MERGEABLE";
    private static final String FIX = "FIX";
    private boolean hasError = false;

    private OperationSwapper() {
    }

    public static AbstractDetector getInstance() {
        return instance;
    }

    @Override
    protected void execute() {
        if (DetectOption.getInstance().isSwap()) {
            BranchRootInfo rootInfo = BranchRootInfo.getInstance();
            BranchInfo a_bInfo = rootInfo.getBranchInfo(MergeType.ACCEPT);
            BranchInfo j_bInfo = rootInfo.getBranchInfo(MergeType.JOIN);
            ConflictInfo cInfo = rootInfo.getConflictInfo();

            List<OperationHunk> a_hunks = createOpertationHunk(a_bInfo, cInfo);
            if (a_hunks != null) {
                ProjectInfo pInfo = a_bInfo.getProjectInfo();
                swapHunk(pInfo, a_hunks);
            }

            List<OperationHunk> j_hunks = createOpertationHunk(j_bInfo, cInfo);
            if (j_hunks != null) {
                ProjectInfo pInfo = j_bInfo.getProjectInfo();
                swapHunk(pInfo, j_hunks);
            }
        }
    }

    private List<OperationHunk> createOpertationHunk(BranchInfo bInfo, ConflictInfo cInfo) {
        if (!cInfo.isConflict()) {
            return null;
        }

        Map<Integer, String> map = new TreeMap<Integer, String>();

        List<BranchJavaElement> confElems = cInfo.getAllBranchJavaElement(bInfo.getType());
        List<String> confElemNames = new ArrayList<>();
        confElems.forEach(elem -> confElemNames.add(elem.getFullName()));

        ProjectInfo pInfo = bInfo.getProjectInfo();
        List<UnifiedOperation> ops = pInfo.getOperations();
        int size = ops.size();
        for (int i = 0; i < size; i++) {
            UnifiedOperation op = ops.get(i);
            if (op.isCommitOpeartion()) {
                map.put(op.getId(), FIX);
            } else if (op.isFileOperation() || op.isFileCloseOperation()) {
                map.put(op.getId(), MERGEABLE);
            } else {
                map.put(op.getId(), UNKNOWN);
            }
        }

        List<BranchJavaElement> elems = bInfo.getAllBranchJavaElement();
        for (BranchJavaElement elem : elems) {
            String name = elem.getFullName();
            ops = elem.getOperations();
            size = ops.size();
            for (int i = 0; i < size; i++) {
                int id = ops.get(i).getId();

                String value = map.get(id);
                if (isUnknownValue(value)) {
                    map.put(id, name);
                }
            }
        }

        List<Integer> mergeId = new ArrayList<>();
        for (Entry<Integer, String> entry : map.entrySet()) {
            int id = entry.getKey();
            String value = map.get(id);
            if (isMergeableValue(value)) {
                mergeId.add(id);
            } else if (!mergeId.isEmpty() && !confElemNames.contains(value)) {
                for (Integer mid : mergeId) {
                    map.put(mid, value);
                }
                mergeId.clear();
            }
        }

        List<OperationHunk> opHunks = new LinkedList<>();
        OperationHunk hunk = null;
        boolean isConflictElement;
        for (Entry<Integer, String> entry : map.entrySet()) {
            int id = entry.getKey();
            String value = map.get(id);

            if (hunk != null && !hunk.name.equals(value)) {
                opHunks.add(hunk);
                isConflictElement = confElemNames.contains(value);
                hunk = new OperationHunk(value, id, isConflictElement);
            } else if (hunk == null) {
                isConflictElement = confElemNames.contains(value);
                hunk = new OperationHunk(value, id, isConflictElement);
            }
            hunk.lastId = id;
        }

        if (hunk != null) {
            opHunks.add(hunk);
        }
        return opHunks;
    }

    private void swapHunk(ProjectInfo pInfo, List<OperationHunk> hunks) {
        boolean swapped;
        do {
            swapped = false;
            int start = 0;
            int size = hunks.size() - 1;

            for (int i = start; i < size; i++) {
                OperationHunk hunk1 = hunks.get(i);
                if (hunk1.isConflictElement) {
                    OperationHunk hunk2 = hunks.get(i + 1);
                    if (!hunk2.isConflictElement) {
                        swapOperation(pInfo, hunk1, hunk2);
                        Collections.swap(hunks, i, i + 1);
                        swapped = true;
                    }
                }
            }
        } while (swapped);

        adaptFileInfoOperation(pInfo);
    }

    private void swapOperation(ProjectInfo pInfo, OperationHunk hunk1, OperationHunk hunk2) {
        int f1 = hunk1.firstId;
        int l1 = hunk1.lastId;
        int f2 = hunk2.firstId;
        int l2 = hunk2.lastId;

        List<UnifiedOperation> ops = pInfo.getOperations();
        for (int i = l1; i >= f1; i--) {
            for (int j = f2; j <= l2; j++) {
                UnifiedOperation op1 = pInfo.getOperationById(i);
                UnifiedOperation op2 = pInfo.getOperationById(j);

                if (RepositoryElementInfoUtil.isPair(op1.getFileInfo(), op2.getFileInfo()) && op1.isNormalOperation() && op2.isNormalOperation()) {
                    NormalOperation nop1 = (NormalOperation) op1.getIOperation();
                    NormalOperation nop2 = (NormalOperation) op2.getIOperation();

                    int s1 = nop1.getStart();
                    int s2 = nop2.getStart();
                    if (s1 != s2) {
                        boolean isForwardOp1 = s1 < s2;
                        if (isForwardOp1) {
                            if (nop1.isInsertion()) {
                                if (nop2.isInsertion()) {
                                    op2.setStart(s2 - op1.maxInsertedTextLength());
                                } else if (nop2.isDeletion()) {
                                    op2.setStart(s2 - op1.maxInsertedTextLength());
                                }
                            } else if (nop1.isDeletion()) {
                                if (nop2.isInsertion()) {
                                    op2.setStart(s2 + op1.maxDeletedOrCopiedTextLength());
                                } else if (nop2.isDeletion()) {
                                    op2.setStart(s2 + op1.maxDeletedOrCopiedTextLength());
                                }
                            }

                            if (nop1.isReplace()) {
                                int iLength = op1.maxInsertedTextLength();
                                int dLength = op1.maxDeletedOrCopiedTextLength();
                                if (iLength >= dLength) {
                                    op2.setStart(op2.getStart() + dLength);
                                } else {
                                    op2.setStart(op2.getStart() - iLength);
                                }
                            }
                        } else {
                            if (nop1.isInsertion()) {
                                if (nop2.isInsertion()) {
                                    op1.setStart(s1 + op2.maxInsertedTextLength());
                                } else if (nop2.isDeletion()) {
                                    op1.setStart(s1 - op2.maxDeletedOrCopiedTextLength());
                                }
                            } else if (nop1.isDeletion()) {
                                if (nop2.isInsertion()) {
                                    op1.setStart(s1 + op2.maxInsertedTextLength());
                                } else if (nop2.isDeletion()) {
                                    op1.setStart(s1 - op2.maxDeletedOrCopiedTextLength());
                                }
                            }

                            if (nop2.isReplace()) {
                                int iLength = op2.maxInsertedTextLength();
                                int dLength = op2.maxDeletedOrCopiedTextLength();
                                if (iLength >= dLength) {
                                    op1.setStart(op1.getStart() - dLength);
                                } else {
                                    op1.setStart(op1.getStart() + iLength);
                                }
                            }
                        }
                    }
                }

                int a = ops.indexOf(op1);
                int b = ops.indexOf(op2);
                Collections.swap(ops, a, b);
            }
        }
    }

    private int searchOperationIndex(List<UnifiedOperation> ops, int id) {
        int idx = 0;
        for (UnifiedOperation op : ops) {
            if (op.getId() == id) {
                return idx;
            }
            idx++;
        }
        return -1;
    }

    private void adaptFileInfoOperation(ProjectInfo pInfo) {
        List<UnifiedOperation> ops = pInfo.getOperations();
        Map<String, List<Integer>> map = new HashMap<String, List<Integer>>();
        for (UnifiedOperation op : ops) {
            String name = op.getFileInfo().getQualifiedName();
            if (map.containsKey(name)) {
                map.get(name).add(op.getId());
            } else {
                List<Integer> ids = new ArrayList<Integer>();
                ids.add(op.getId());
                map.put(name, ids);
            }
        }

        List<FileInfo> fInfos = pInfo.getAllFileInfo();
        for (FileInfo fInfo : fInfos) {
            List<Integer> ids = map.get(fInfo.getQualifiedName());
            List<UnifiedOperation> new_ops = new ArrayList<>();
            for (int id : ids) {
                int idx = searchOperationIndex(ops, id);
                new_ops.add(ops.get(idx));
            }
            fInfo.setOperations(new_ops);
        }
    }

    private boolean isUnknownValue(String value) {
        return value.equals(UNKNOWN);
    }

    private boolean isMergeableValue(String value) {
        return value.equals(MERGEABLE);
    }

    @Override
    protected String getTaskName() {
        return TASK_NAME;
    }

    @Override
    protected String getErrorMessage() {
        return ERROR_MESSAGE;
    }

    @Override
    protected void nextState() {
        if (!hasError) {
            AMergePointSearcher.getInstance().detect(dialog);
        } else {
            error();
        }
    }

    private class OperationHunk {
        String name;
        int firstId, lastId;
        boolean isConflictElement;

        OperationHunk(String name, int firstId, boolean isConflictElement) {
            this.name = name;
            this.firstId = firstId;
            this.isConflictElement = isConflictElement;
        }
    }
}
